; Nametable "D" Compression decoder
; Copyright (c) 2015-2016  Johnathan Roatch
;
; Copying and distribution of this file, with or without
; modification, are permitted in any medium without royalty provided
; the copyright notice and this notice are preserved in all source
; code copies.  This file is offered as-is, without any warranty.
;
; Version History:
;   2016-09-08: The highest bit of PPU_ADDR also sets PPU_CTRL 32 INC mode.
;               All this new stuff is now optional, as there are better ways
;               to composite different types of screen data.
;               (bytes: 98 ROM, 5 RAM)
;   2016-09-06: Made 'set PPU_ADDR' reset the BG byte.
;               (bytes: 135 ROM, 5 RAM)
;   2016-09-05: Removed pb8 mode. Optimized format, by trading some
;               'runs' speed for ROM space. Added signaled exit.
;               (bytes: 117 ROM, 5 RAM)
;   2016-01-02: Small optimization in control byte decode.
;               (bytes: 156 ROM, 5 RAM)
;   2015-12-31: Format rearranged, renamed to PBJ.
;               (bytes: 159 ROM, 5 RAM)
;   2015-11-21: Removed "transperency" and added a pb8 mode
;               to help with tile compression.
;               (bytes: 172 ROM, 5 RAM)
;   2015-09-20: Merged lite variant to use compile-time options instead.
;               Optimized PPU_ADDR command by 2 btyes.
;               (bytes: 105 ROM, 4 RAM) (PPU_ADDR: +14 ROM)
;   2015-09-16: Optimized away another temporary in double byte mode.
;               (bytes: 146 ROM, 5 RAM)
;   2015-03-09: Changed meaning of 000nnnnn to "32-N literal bytes."
;               Added PPU_ADDR commands.
;               Separated into a lite variant.
;               (bytes: 151 ROM, 6 RAM)
;   2015-03-02: Optimized away a temporary in double byte mode.
;               (bytes: 118 ROM, 5 RAM)
;   2015-03-01: Initial release, (bytes: 123 ROM, 6 RAM)
;
; Stream format, (with all compile options enabled):
;     0nnnnnnn        : 128-N Run of BG.
;     100nnnnn ...    : 32-N literal bytes.
;     10111111 xx     : set BG to X. (initially BG = 0)
;     10111110 xx yy  : Set PPU_ADDR to yyxx, and sets BG to 0.
;                       if yyxx >= $8000, then set PPU_CTRL to 32 INC mode.
;     101nnnnn xx yy  : For 32-N times, write alternately X and Y.
;     11011111        : End stream with V flag cleared.
;     110nnnnn xx     : 32-N incrementing run starting at X.
;     11111111        : End stream with V flag set.
;     111nnnnn xx     : 32-N run of X.
;

.include "nes.inc"

temp = $00

.export ntd_decode
.importzp ntd_stream_ptr


.segment "CODE"

; Compile time options. base amount of space: 96 ROM, 5 RAM

; I go out the way to make $ff the end command at
; the expense of an extra `ora`, so that cleared flash
; will end a runaway stream. The encoder assumes this is set.
.ifndef NTD_INVERSE_BG_RUN_BIT
NTD_INVERSE_BG_RUN_BIT = 1  ; (bytes: +2 ROM, +0 RAM)
.endif

; if enabled, the $be command will read 2 bytes to set PPU_ADDR with.
; The $be command will also set the BG byte back to 0.
.ifndef NTD_USE_PPUADDR_SET
NTD_USE_PPUADDR_SET = 0     ; (bytes: +14 ROM, +0 RAM)
.endif

; if enabled. Then the $df command will be the cleared V flag variant
; of the $ff end command.
.ifndef NTD_USE_V_EXIT_SIGNAL
NTD_USE_V_EXIT_SIGNAL = 0     ; (bytes: +4 ROM, +0 RAM)
.endif

; This is an experimental hack to see how well this would work as a
; generic PPU string thing. Enabling this will cause PPU_CTRL to be always used.
.if NTD_USE_PPUADDR_SET
.ifndef NTD_USE_VRAM_DOWN_OF_PPUCTRL
NTD_USE_VRAM_DOWN_OF_PPUCTRL = 0     ; (bytes: +18 ROM, +0 RAM)
.endif
.else
NTD_USE_VRAM_DOWN_OF_PPUCTRL = 0
.endif

;;
; decodes the stream directly to the PPU_DATA
;
; Input: start of stream pointer. PPU_ADDR is assumed set.
; Output: stream pointer points to after the end.
; Trashes: A,X,Y, temp 0 ~ 2.
.proc ntd_decode
stream_ptr      = ntd_stream_ptr
stream_ptr_lo   = stream_ptr+0
stream_ptr_hi   = stream_ptr+1
current_action  = temp+0
bg_byte         = temp+1
dbl_byte        = temp+2
  ldx #$00
  ldy stream_ptr_lo
  stx stream_ptr_lo
.if ::NTD_USE_VRAM_DOWN_OF_PPUCTRL
  clv
reset_bg_byte:
  lda #VBLANK_NMI
  bvc dont_paint_down
    lda #VBLANK_NMI|VRAM_DOWN
  dont_paint_down:
  sta PPU_CTRL
.else
reset_bg_byte:
.endif
  txa
  dex
set_bg_byte:
  sta bg_byte
  inx
  input_loop:
    lda (stream_ptr), y
    iny
    bne @__skip1
      inc stream_ptr_hi
    @__skip1:
    cpx #$00
    beq decode_control_byte
      bit current_action  ;N = b7, V = b6
      bmi run_or_inc
      lit_or_dbl:
        bvs do_dbl
        do_lit:  ; 100nnnnn
          sta PPU_DATA
          inx
        jmp input_loop
        do_dbl:  ; 111nnnnn
          cpx #$ff
          beq set_bg_byte
          eor (stream_ptr), y
          sta dbl_byte
          lda (stream_ptr), y
.if ::NTD_USE_VRAM_DOWN_OF_PPUCTRL
          bmi leave_V_set
             clv
          leave_V_set:
.endif
          iny
          bne @__skip2
            inc stream_ptr_hi
          @__skip2:
.if ::NTD_USE_PPUADDR_SET
          cpx #$fe
          bne do_run
        set_ppu_addr:
          set_ppu_addr_loop:
            sta PPU_ADDR
            eor dbl_byte
            inx
          bne set_ppu_addr_loop
        beq reset_bg_byte   ; bra
.endif
        ;,; bvs do_run
      run_or_inc:
        bvs do_run
        do_inc: ; 101nnnnn
          inc_loop:
            sta PPU_DATA
            clc
            adc #$01
            inx
          bne inc_loop
        beq input_loop   ; bra
    decode_control_byte:
      stx dbl_byte
      sta current_action
      asl current_action
.if ::NTD_INVERSE_BG_RUN_BIT
      bcs not_bgrun
        ora #%10000000
.else
      bcc not_bgrun
.endif
        do_bgrun:  ; 0nnnnnnn
          tax
          lda bg_byte
        do_run:  ; 110nnnnn
          run_loop:
            eor dbl_byte
            sta PPU_DATA
            inx
          bne run_loop
        beq input_loop   ; bra
      not_bgrun:
.if ::NTD_USE_V_EXIT_SIGNAL
      ora #%00100000
.endif
.if ::NTD_INVERSE_BG_RUN_BIT
      cmp #%11111111
.else
      cmp #%01111111
.endif
      beq end_stream
      ora #%11100000
      tax
    bne input_loop   ; bra
end_stream:
.if ::NTD_USE_VRAM_DOWN_OF_PPUCTRL
  lda #VBLANK_NMI
  sta PPU_CTRL
.endif
.if ::NTD_USE_V_EXIT_SIGNAL
  bit current_action
.endif
  sty stream_ptr_lo
rts
.endproc



