; NES controller reading code
; Copyright 2016 Johnathan Roatch
;

.include "nes.inc"

.importzp temp
.export pads_read, pads_rejectOpposingDirections

.segment "CODE"

;;
; Read 16 bits from both NES controller ports MSB first but in little endian
; byte order. Also keeps in mind the even odd cycles after a DMA aliment.
; To avoid DMC glitches, call this immediately after the `sta OAM_DMA`.
; http://forums.nesdev.com/viewtopic.php?p=172188#p172188
;
; Preserves X, Y.
.proc pads_read
pad_1_keys = temp+0
pad_1_axlr = temp+1
pad_2_keys = temp+2
pad_2_axlr = temp+3
    ; 16 Bits from the controllers are shifted into pad_1 and pad_2.
    ; In addition, $0001 in pad_2 serves as the loop counter:
    ; once the bit gets shifted left sixteen times, it'll end up in carry
    ; terminating the read loop.
    ;
    ; Also the strobe part must take an odd total number of cycles.
  lda #$01
  sta pad_2_axlr
    ; Write 1 then 0 to JOY_PORT_1 to send a latch signal, telling the
    ; controllers to copy button states into a shift register.
  sta JOY_STROBE
    ; We need an extra cycle here that functions the same as `lda #0`
    ; XORing with itself will clear all the bits in A.
  eor pad_2_axlr
  sta pad_2_keys
  sta JOY_STROBE
    ; The loop part must take an even number of cycles.
  @read_loop:
      ; On NES and AV Famicom, button presses always show up in D0.
      ; On the original Famicom, presses on the hardwired controllers
      ; show up in D0 and presses on plug-in controllers show up in D1.
      ; D2-D7 consist of data from the Zapper, Power Pad, Vs. System
      ; DIP switches, and bus capacitance; so we ignore them.
    lda JOY_PORT_1    ; read player 1's controller
    and #%00000011    ; ignore D2-D7
    cmp #1            ; CLC if A=0, SEC if A>=1
    rol pad_1_axlr    ; shift the the bit into memory
    rol pad_1_keys
    lda JOY_PORT_2    ; read player 2's controller the same way
    and #%00000011
    cmp #1
    rol pad_2_axlr
    rol pad_2_keys
      ; Even out the odd cycles in the loop branch.
      ; 3 cycles with harmless flag setting side effects.
    bit $24
  bcc @read_loop
    ; once $0001 has been shifted 16 times, we're done
rts
.endproc

.proc pads_rejectOpposingDirections
pad_1_keys = temp+0
pad_1_axlr = temp+1
pad_2_keys = temp+2
pad_2_axlr = temp+3
  ldx #temp+0
  jsr fix_byte
  ldx #temp+2
fix_byte:
  ; Construct a XOR mask for when up+down or left+right are pressed.
  ; First apply bitwise AND to up+down and left+right.
  lda 0,x
  asl
  and 0,x
  ; mask out other buttons.
  and #%00001010
  ; Then dupplicate the resulting bits to each direction.
  sta temp+4
  lsr
  ora temp+4
  ; If all directions are pressed simultaneously, then clear all keys.
  ; this is to prevent loose connectors from accidentally soft reseting,
  ; and so that "The Game Handler" won't be held upside-down.
  cmp #BUTTON_UP|BUTTON_DOWN|BUTTON_LEFT|BUTTON_RIGHT
  bne @dont_clear_all_keys
    ; loading the same value as the value to XOR with
    ; so that the result is $00 being stored to 0,x
    lda 0,x
  @dont_clear_all_keys:
  ; Constructed mask now contains the buttons to toggle off.
  eor 0,x
  sta 0,x
rts
.endproc

