#!/usr/bin/env python3
# "Bagel", NES CHR compressor,
# Copyright (C) 2015  Johnathan Roatch
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import random

def pb8_pack_plane(plane, top_value=None, prev_plane=None):
    plane_iter = iter(plane)
    cplane = bytearray(1)
    flags = 0
    cur_byte = top_value
    for i in range(8):
        if prev_plane:
            prev_byte = prev_plane[i]
        else:
            prev_byte = cur_byte
        cur_byte = next(plane_iter)
        flags = flags << 1
        if cur_byte != prev_byte:
            flags = flags | 1
            cplane.append(cur_byte)
    cplane[0] = flags
    return bytes(cplane)

def pb8_unpack_plane(cplane, top_value=0x00, prev_plane=None, flags=None):
    cplane_iter = iter(cplane)
    plane = bytearray(8)
    if flags is None:
        flags = next(cplane_iter)
    cur_byte = top_value
    for i in range(8):
        flags = flags << 1
        if flags & 0x0100:
            cur_byte = next(cplane_iter)
        elif prev_plane:
            cur_byte = prev_plane[i]
        plane[i] = cur_byte
    return bytes(plane)

def compress_block(block, prev_block=None, randomize_equivalent_choices=False):
    block_iter = iter(block)
    block_bytes = bytes( [next(block_iter) for _ in range(64)] )
    if block_bytes == prev_block:
        if randomize_equivalent_choices:
            return bytes([0xcc | random.randrange(256)])
        else:
            return b'\xff'
    block_plane_choices = []
    prev_plane = None
    inv_prev_plane = None
    block_short_flag_choices = { i|0xf0 for i in range(16) }
    prev_plane_flag_choices = []
    for i in range(8):
        plane_choices = []
        plane = block_bytes[i*8:i*8+8]
        if plane == b'\x00'*8:
            plane_choices.append( (0b00, b'') )
        if not plane_choices:
            if prev_block and plane == prev_block[i*8:i*8+8]:
                plane_choices.append( (0b10, b'\x00') )
            if plane == b'\xff'*8:
                plane_choices.append( (0b10, b'\xff') )
            if prev_plane and plane == prev_plane:
                plane_choices.append( (0b11, b'\x00') )
            if inv_prev_plane and plane == inv_prev_plane:
                plane_choices.append( (0b11, b'\xff') )
        if not plane_choices:
            cplane = pb8_pack_plane(plane, top_value=0x00)
            if prev_plane:
                cplane_prev = pb8_pack_plane(plane, prev_plane=prev_plane)
            else:
                cplane_prev = bytes(9)
            if 1 < len(cplane) < 9 and len(cplane) <= len(cplane_prev):
                plane_choices.append( (0b10, cplane) )
            if 1 < len(cplane_prev) < 9 and len(cplane_prev) <= len(cplane):
                plane_choices.append( (0b11, cplane_prev) )
        if not plane_choices or len(plane_choices[0][1]) >= 8:
            plane_choices.append( (0b01, plane) )
        block_plane_choices.append(plane_choices)

        plane_flag_choices = [_[0] for _ in plane_choices]
        if i & 1 == 1:
            t = set()
            for y in prev_plane_flag_choices:
                for x in plane_flag_choices:
                    t.add( y<<2 | x | 0xf0 )
            block_short_flag_choices &= t
        prev_plane_flag_choices = plane_flag_choices

        prev_plane = plane
        inv_prev_plane = bytes( [_^0xff for _ in plane] )
    block = [b'']
    if block_short_flag_choices:
        if randomize_equivalent_choices:
            short_flags = random.choice(list(block_short_flag_choices))
            block[0] = bytes([short_flags ^ (random.randrange(256)&0x30)])
        else:
            short_flags = min(block_short_flag_choices)
            block[0] = bytes([short_flags])
        for i in range(8):
            plane_choices = block_plane_choices[i]
            if i & 1 == 1:
                plane_mode = short_flags & 0b11
            else:
                plane_mode = short_flags>>2 & 0b11
            if randomize_equivalent_choices:
                plane_choice = random.choice([_ for _ in plane_choices if _[0] == plane_mode])
            else:
                plane_choice = [_ for _ in plane_choices if _[0] == plane_mode][0]
            block.append(plane_choice[1])
    else:
        block_flags = 0x0000
        for i in block_plane_choices:
            block_flags = block_flags << 2
            if randomize_equivalent_choices:
                plane_choice = random.choice(i)
            else:
                plane_choice = i[0]
            block_flags = block_flags | plane_choice[0]
            block.append(plane_choice[1])
        block[0] = bytes([block_flags>>8 & 0xff, block_flags & 0xff])
    return b''.join(block)

def decompress_block(cblock, prev_block=None):
    cblock_iter = iter(cblock)
    block = bytearray(64)
    block_flags = next(cblock_iter)
    if block_flags >= 0xc0:
        i = block_flags & 0x0f
        block_flags = i | i<<4 | i<<8 | i<<12
    else:
        block_flags = block_flags << 8 | next(cblock_iter)
    if block_flags>>8 & 0xff >= 0xc0:
        if prev_block is None:
            raise ValueError('block referenced non-existent previous block.')
        return bytes(prev_block)
    prev_plane = None
    for i in range(8):
        block_flags = block_flags<<2
        plane_mode = (block_flags>>16)&0b11
        if plane_mode == 0b00:
            new_plane = b'\x00'*8
        elif plane_mode == 0b01:
            new_plane = bytes( [next(cblock_iter) for _ in range(8)] )
        elif plane_mode == 0b10:
            plane_flags = next(cblock_iter)
            if plane_flags == 0xff:
                new_plane = b'\xff'*8
            elif plane_flags == 0x00:
                if prev_block is None:
                    raise ValueError('block referenced non-existent previous block.')
                new_plane = prev_block[i*8:i*8+8]
            else:
                new_plane = pb8_unpack_plane(cblock_iter, top_value=0x00, flags=plane_flags)
        elif plane_mode == 0b11:
            plane_flags = next(cblock_iter)
            if plane_flags == 0xff:
                new_plane = bytes( [_^0xff for _ in prev_plane] )
            else:
                new_plane = pb8_unpack_plane(cblock_iter, prev_plane=prev_plane, flags=plane_flags)
        block[i*8:i*8+8] = new_plane
        prev_plane = new_plane
    return bytes(block)

# TODO: optimize and use this.
def fill_partial_block(block, prev_block=None):
    block_length = len(block)
    if block_length == 64:
        return block
    elif block_length > 64:
        return block[0:64]
    if prev_block and block == prev_block[0:block_length]:
        return prev_block
    complete_planes = block[0:block_length&0x78]
    partial_plane = block[block_length&0x78:]
    if partial_plane:
        partial_plane_length = len(partial_plane)
        last_byte = partial_plane[-1]
        trail_plane = partial_plane + bytes([last_byte]*(8-partial_plane_length))
        trail = pb8_pack_plane(trail_plane, top_value=0x00)
        if trail == b'\x80\xff':
            trail = b'\xff'
        if prev_block and len(trail) > 1:
            if partial_plane == prev_block[block_length&0x78:block_length]:
                trail_plane = prev_block[block_length&0x78:(block_length&0x78)+8]
                trail = b'\x00'
        if complete_planes and len(trail) > 1:
            trail_prev_plane = partial_plane + complete_planes[-8+partial_plane_length:]
            trail_prev = pb8_pack_plane(trail_prev_plane, prev_plane=complete_planes[-8:])
            if len(trail_prev) < len(trail):
                trail = trail_prev
                trail_plane = trail_prev_plane
            else:
                trail_inv_prev_plane = bytes([i^0xff for i in complete_planes[-8:]])
                if partial_plane == trail_inv_prev_plane[0:partial_plane_length]:
                    trail = b'\xff'
                    trail_plane = trail_inv_prev_plane
        partial_plane = trail_plane
    zero_planes_length = 64 - (len(complete_planes) + len(partial_plane))
    return complete_planes + partial_plane + bytes(zero_planes_length)

# TODO: implment these methods
#def compress(data, prepend_number_of_blocks=True, try_interleaving=True, forbid_block_references=False):
#    # With prepend_number_of_blocks: 
#    #   When try_interleaving is True, the max block count is 128
#    #   otherwise it's 192
#    pass
#
#def decompress(data, number_of_blocks=True, allow_interleaving=True, force_interleaving=False):
#    pass

if __name__=='__main__':
    def main_encode_blocks(chr_sheet, swizzle=False):
        coded_blocks = []
        number_of_blocks = len(chr_sheet)//64
        prev_block = None
        for block_number in range(number_of_blocks):
            if swizzle:
                i = (block_number//2) | ((block_number&0x1)*64)
                block = chr_sheet[i*64:i*64+64]
            else:
                block = chr_sheet[block_number*64:block_number*64+64]
            cblock = compress_block(block, prev_block=prev_block)
            #if decompress_block(cblock, prev_block=prev_block) != block:
            #    print("Encoding error!", (block, prev_block, cblock))
            coded_blocks.append(cblock)
            prev_block = block
        return b''.join(coded_blocks)

    def main_decode_blocks(file_iter, number_of_blocks, swizzle=False):
        chr_sheet = bytearray(number_of_blocks*64)
        prev_block = None
        for block_number in range(number_of_blocks):
            block = decompress_block(file_iter, prev_block=prev_block)
            if swizzle:
                i = (block_number//2) | ((block_number&0x1)*64)
                chr_sheet[i*64:i*64+64] = block
            else:
                chr_sheet[block_number*64:block_number*64+64] = block
            prev_block = block
        return bytes(chr_sheet)

    import sys
    import argparse
    import itertools
    import functools

    parser = argparse.ArgumentParser(description='NES Bagel Compressor', usage='%(prog)s [-d] [options] -o output [-i] input')
    parser.add_argument('--version', action='version', version='%(prog)s 2.0')
    parser.add_argument('input', metavar='files', help='Input files', nargs='*')
    parser.add_argument('-o', '--output', metavar='file', help='Output file')
    parser.add_argument('-f', '--force',  help='Force overwrite of output file.', action='store_true')
    parser.add_argument('-d', '--decompress', help='Decode the input file.', action='store_true')
    options = parser.parse_args()

    if '-' not in options.input and not sys.stdin.isatty():
        options.input.append('-')

    if options.output is None and not sys.stdout.isatty():
        options.output = '-'

    if not options.input or not options.output:
        sys.exit(parser.format_help())

    if options.output == '-':
        output_file = sys.stdout.buffer
    elif options.force:
        output_file = open(options.output, 'wb')
    else:
        output_file = open(options.output, 'xb')

    for fn in options.input:
        if fn == '-':
            input_file = sys.stdin.buffer
        else:
            input_file = open(fn, 'rb')

        if options.decompress:
            in_file_iter = itertools.chain.from_iterable(
                iter(functools.partial(input_file.read, 8192), b'')
            )
            for number_of_blocks in in_file_iter:
                if number_of_blocks == 0:
                    chr_sheet = main_decode_blocks(in_file_iter, 128, swizzle=True)
                else:
                    chr_sheet = main_decode_blocks(in_file_iter, number_of_blocks, swizzle=False)
                output_file.write(chr_sheet)
            #TODO: catch partial block groups, and partial blocks.
        else:
            chr_sheet = input_file.read(8192)
            number_of_blocks = len(chr_sheet)//64
            while number_of_blocks > 0:
                coded_sheet = main_encode_blocks(chr_sheet, swizzle=False)
                if len(chr_sheet) == 8192:
                    coded_sheet_swizzle = main_encode_blocks(chr_sheet, swizzle=True)
                    if len(coded_sheet_swizzle) < len(coded_sheet):
                        coded_sheet = coded_sheet_swizzle
                        number_of_blocks = 0  # special tag
                output_file.write(bytes([number_of_blocks]))
                output_file.write(coded_sheet)
                chr_sheet = input_file.read(8192)
                number_of_blocks = len(chr_sheet)//64

        if fn != '-':
            input_file.close()

    if options.output != '-':
        output_file.close()

#def test_generate_random_decompress():
#    import os
#    blocks = []
#    cblocks = []
#    random_bytes = os.urandom(64*66)
#    random_bytes_iter = iter(random_bytes)
#    prev_block = bytes( (random_bytes[i]^0xa5 for i in range(64)) )
#    for i in range(64):
#        this_block = decompress_block(random_bytes_iter, prev_block=prev_block)
#        blocks.append(this_block)
#        this_cblock = compress_block(this_block, prev_block=prev_block, randomize_equivalent_choices=True)
#        cblocks.append(this_cblock)
#        prev_block = this_block
#    return [cblocks, blocks]
#
#def test_generate():
#    chr_data_input = b''
#    with open("test-bitmap.chr", 'rb') as input_file:
#        chr_data_input = input_file.read(4096)
#    input_byte_iter = iter(chr_data_input)
#    cblocks = [bytes([128])]
#    prev_block = None
#    for i in range(len(chr_data_input)//64):
#        this_cblock = compress_block(input_byte_iter, prev_block=prev_block, randomize_equivalent_choices=True)
#        cblocks.append(this_cblock)
#        if i < 4:
#            print( (chr_data_input[i*64:i*64+64], prev_block, this_cblock) )
#        prev_block = chr_data_input[i*64:i*64+64]
#    bitmap_stream_length = sum( (len(i) for i in cblocks) )
#
#    stream_length = bitmap_stream_length
#    while stream_length != 4096:
#        trail_random_run = test_generate_random_decompress()
#        random_stream_length = sum( (len(i) for i in trail_random_run[0]) )
#        stream_length = bitmap_stream_length + random_stream_length
#        print(stream_length)
#
#    with open("test-stream.chr", 'wb') as output_chr_file:
#        output_chr_file.write(chr_data_input)
#        output_chr_file.write(b''.join(trail_random_run[1]))
#    with open("test-stream.chr.bagel", 'wb') as output_file:
#        output_file.write(b''.join(cblocks))
#        output_file.write(b''.join(trail_random_run[0]))
#    return
